/*
 * mkTulip 1.0
 * -----------
 *
 * Please see the web pages for releases and documentation.
 *
 * Author: Michael Kraus
 *         mailto:mmg_kraus@csi.com
 *         http://ourworld.compuserve.com/homepages/mmg_kraus
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted.
 * The author makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express
 * or implied warranty.  By use of this software the user agrees to
 * indemnify and hold harmless the author from any claims or
 * liability for loss arising out of such use.
 *
 */

/* required to built a dll using stubs. should be a compiler option */
/* #define USE_TCL_STUBS */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <tcl.h>

#ifndef TRUE
#  define TRUE  1
#  define FALSE 0
#endif

/* copied from sun's example.c */
#ifdef __WIN32__
#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN
#   if defined(_MSC_VER)
#       define EXPORT(a,b) __declspec(dllexport) a b
#       define DllEntryPoint DllMain
#   else
#       if defined(__BORLANDC__)
#           define EXPORT(a,b) a _export b
#       else
#           define EXPORT(a,b) a b
#       endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif

EXTERN EXPORT(int,Mktulip_Init) _ANSI_ARGS_((Tcl_Interp *interp));
EXTERN EXPORT(int,Mktulip_SafeInit) _ANSI_ARGS_((Tcl_Interp *interp));
BOOL APIENTRY DllEntryPoint(HINSTANCE hInst, DWORD reason, LPVOID reserved)
{
    return TRUE;
}
#endif

/* mkTulip version number */
#define _VERSION           "1.0"

/* buffer length = max. # of digits + 1 */
#define _BUFLEN            1024

/* some acronyms for popular Tcl_xxx functions */
#define _NSO(pcText)       Tcl_NewStringObj( pcText, -1 )
#define _SSO(pO,pcText)    Tcl_SetStringObj( pO, pcText, -1 )
#define _GSO(pO)           Tcl_GetStringFromObj( pO, NULL )

#define _NIO(iVal)         Tcl_NewIntObj( iVal )
#define _SIO(pO,iVal)      Tcl_SetIntObj( pO, iVal )
#define _GIO(pO,piVal)     Tcl_GetIntFromObj( pI, pO, piVal )

#define _NBO(bVal)         Tcl_NewBooleanObj( bVal )
#define _SBO(pO,bVal)      Tcl_SetBooleanObj( pO, bVal )
#define _GBO(pO,pbVal)     Tcl_GetBooleanFromObj( pI, pO, pbVal )

#define _NDO(fVal)         Tcl_NewDoubleObj( fVal )
#define _SDO(pO,fVal)      Tcl_SetDoubleObj( pO, fVal )
#define _GDO(pO,pfVal)     Tcl_GetDoubleFromObj( pI, pO, pfVal )

#define _NLO(iC,ppO)       Tcl_NewListObj( iC, ppO )

#define _LOAL(pO,pNewO)    Tcl_ListObjAppendList( pI, pO, pNewO )
#define _LOAE(pO,pNewO)    Tcl_ListObjAppendElement( pI, pO, pNewO )
#define _LOGL(pO,piLen)    Tcl_ListObjLength( pI, pO, piLen )
#define _LOGI(pO,iI,poE)   Tcl_ListObjIndex( pI, pO, iI, poE )
#define _LOGE(pO,piC,ppV)  Tcl_ListObjGetElements( pI, pO, piC, ppV )
#define _LORE(pO,iPos,poE) Tcl_ListObjReplace( pI, pO, iPos, 1, 1, poE )
#define _LODE(pO,iPos)     Tcl_ListObjReplace( pI, pO, iPos, 1, 0, NULL )
#define _LODR(pO,iP1,iP2)  Tcl_ListObjReplace( pI, pO, iP1, (iP2)-(iP1)+1, 0, NULL )
#define _LOIE(pO,iPos,poE) Tcl_ListObjReplace( pI, pO, iPos, 0, 1, poE )
#define _LOIR(pO,iPos,iC,poE) Tcl_ListObjReplace( pI, pO, iPos, 0, iC, poE )
#define _LORR(pO,iP1,iP2,iC,ppV)  Tcl_ListObjReplace( pI, pO, iP1, (iP2)-(iP1)+1, iC, ppV )

#define _NOB               Tcl_NewObj()
#define _DOB               Tcl_DuplicateObj
#define _SOB(pO)           Tcl_IsShared( pO )? Tcl_DuplicateObj( pO ):pO
#define _ASO               Tcl_AppendStringsToObj

#define _DRC               Tcl_DecrRefCount
#define _IRC               Tcl_IncrRefCount

#define _GOR               Tcl_GetObjResult( pI )
#define _SOR(pO)           Tcl_SetObjResult( pI, pO )
#define _ROR               Tcl_ResetResult( pI )

#define _GIFO(pO,pA,pcTxt,piRes) Tcl_GetIndexFromObj( pI, pO, pA, pcTxt, 0, piRes )
#define _WNA(objc,pcText)        ( Tcl_WrongNumArgs( pI, objc, objv, pcText ), TCL_ERROR )

#define try( Expr, Excep ) { if( Expr != TCL_OK ) throw Excep; }
#define throw              goto
#define catch

/* struct for long integers */
typedef struct _MkuValue {
  char pcAbs[_BUFLEN];                      /* value without zeros and sign */
  int  iLen;                                    /* length of value in bytes */
  int  bNeg;                                   /* true if value is negative */
} _MkuValue;

/* function prototypes. */
int Mktulip_Init( Tcl_Interp * );
int Mktulip_SafeInit( Tcl_Interp * );
int Mku_TulipCmd ( ClientData, Tcl_Interp *, int, Tcl_Obj *CONST[] );

/* _MkuError
   creates a formatted result string and always returns TCL_ERROR.
   like with printf(), at least a format string must be provided.
   in addition, the format string may contain "%O" for Tcl_Obj* types.
*/
static int _MkuError( Tcl_Interp *pI, char *pcFormat, ... )
{
  int     i;
  char  **args, *pcRun, pcMsg[2000], pcFmt[2000];
  va_list marker;

  if( pI == NULL || pcFormat == NULL )
    return TCL_ERROR;

  va_start( marker, pcFormat );
  args = (char**)marker;

  strcpy( pcFmt, pcFormat );

  for( i = 0, pcRun = pcFmt; *pcRun; pcRun++ )
  {
    if( *pcRun != '%' ) continue;

    pcRun++;
    if( *pcRun == 'O' )
    {
      *pcRun = 's';
      args[i] = _GSO( (Tcl_Obj*)args[i] );
      i++;
    }
  }

  vsprintf( pcMsg, pcFmt, marker );
  _SSO( _GOR, pcMsg );

  return TCL_ERROR;
}

/* _MkuParse
   tries to convert a string that is expected to be an integer into an
   _MkuValue structure. a leading + or - sign as well as leading zeros
   are cut off. if the remaining string contains chars other than 0-9,
   or if the string is empy, then the function returns 1, else 0.
*/
int _MkuParse( char *pcV, _MkuValue *psVal )
{
  char *pcC, *pcR = pcV;

  if( ! *pcV )
    return 1;                            /* complain if its an empty string */

  psVal->bNeg = ( *pcR == '-' );                         /* detect '-' sign */

  if( psVal->bNeg || *pcR == '+' )         /* jump over any '+' or '-' sign */
    pcR++;

  while( *pcR == '0' && *(pcR+1) )              /* cut off leading zeros... */
    pcR++;                                  /* but not if its the last char */

  if( *pcR == '0' )                     /* if all that is left is a zero... */
    psVal->bNeg = 0;                     /* then make sure sign is positive */

  for( pcC = pcR; *pcC; pcC++ )             /* check for invalid characters */
    if( *pcC < '0' || *pcC > '9' )
      return 1;                       /* complain if a non-numeric is found */

  strcpy( psVal->pcAbs, pcR );                /* finally copy trimmed value */
  psVal->iLen = pcC - pcR;                          /* and store its length */

  return 0;
}

/* _MkuCmp
   compares two values and determines which one is greater or if they are
   equal. -1 is returned if *psV1 < *psV2, 1 if psV1 > psV2, 0 if both
   are equal.
*/
int _MkuCmp( _MkuValue *psV1, _MkuValue *psV2 )
{
  int iCmp;

  /* simply check if signs are different */
  if( psV1->bNeg && ! psV2->bNeg )
    return -1;
  else if( ! psV1->bNeg && psV2->bNeg )
    return 1;

  /* signs are identical. now check if one value is shorter */
  if( psV1->iLen < psV2->iLen )
    return psV1->bNeg ? 1 : -1;
  else if( psV1->iLen > psV2->iLen )
    return psV1->bNeg ? -1 : 1;

  /* both values are of equal length and have identical signs */
  /* alphanumeric comparison is o.k. now, but limit it to -1..1 */
  iCmp = strcmp( psV1->pcAbs, psV2->pcAbs );
  if( iCmp >  1 ) iCmp =  1;
  if( iCmp < -1 ) iCmp = -1;

  if( psV1->bNeg )
    return -iCmp;
  else
    return iCmp;
}

/* _MkuCmpAbs
   same as _MkuCmp, only that the signs of the two numbers are ignored.
*/
int _MkuCmpAbs( _MkuValue *psV1, _MkuValue *psV2 )
{
  int iCmp;

  if( psV1->iLen < psV2->iLen )
    return -1;
  else if( psV1->iLen > psV2->iLen )
    return 1;

  iCmp = strcmp( psV1->pcAbs, psV2->pcAbs );
  if( iCmp >  1 ) iCmp =  1;
  if( iCmp < -1 ) iCmp = -1;

  return iCmp;
}

/* _MkuAdd
   substraction routine. psV1 must be greator or equal to psV2. also,
   the function does not look at the signs of the two values, nor does
   it set the sign of the result. must be done by the caller.
*/
void _MkuAdd( _MkuValue *psV1, _MkuValue *psV2, _MkuValue *psR )
{
  char *pcV1, *pcV2, *pcR, pcBuf[_BUFLEN];
  int iL1, iL2, iS, iC;

  pcV1 = psV1->pcAbs; iL1 = psV1->iLen;
  pcV2 = psV2->pcAbs; iL2 = psV2->iLen;

  pcR = pcBuf + _BUFLEN - 1;
  *pcR = 0;
  iC = 0;

  while( iL1 || iL2 )
  {
    if( iL1 && iL2 )
      iS = ( ( pcV1[--iL1] - '0' ) + ( pcV2[--iL2] - '0' ) + iC );
    else if( iL1 )
      iS = ( ( pcV1[--iL1] - '0' ) + iC );
    else if( iL2 )
      iS = ( ( pcV2[--iL2] - '0' ) + iC );

    iC = iS / 10;
    *--pcR = ( iS % 10 ) + '0';
  }

  if( iC )
    *--pcR = iC + '0';

  strcpy( psR->pcAbs, pcR );
  psR->iLen = strlen( psR->pcAbs );
}

/* _MkuSub
   substraction routine. psV1 must be greator or equal to psV2. also,
   the function does not look at the signs of the two values, nor does
   it set the sign of the result. must be done by the caller.
*/
void _MkuSub( _MkuValue *psV1, _MkuValue *psV2, _MkuValue *psR )
{
  char *pcV1, *pcV2, *pcR, pcBuf[_BUFLEN];
  int iL1, iL2, iS, iC;

  pcV1 = psV1->pcAbs; iL1 = psV1->iLen;
  pcV2 = psV2->pcAbs; iL2 = psV2->iLen;

  pcR = pcBuf + _BUFLEN - 1;
  *pcR = 0;
  iC = 0;

  while( iL1 || iL2 )
  {
    if( iL1 && iL2 )
      iS = ( ( pcV1[--iL1] - '0' ) - ( pcV2[--iL2] - '0' ) - iC + 10 );
    else if( iL1 )
      iS = ( ( pcV1[--iL1] - '0' ) - iC + 10 );

    iC = 1 - iS / 10;
    *--pcR = ( iS % 10 ) + '0';
  }

  while( *pcR == '0' && *(pcR+1) )
    pcR++;

  strcpy( psR->pcAbs, pcR );
  psR->iLen = strlen( psR->pcAbs );
}

/* _MkuMul
   multiplication routine. just like in school. it's a series of additions.
   the function does not look at the signs of the two values, nor does
   it set the sign of the result. must be done by the caller.
*/
void _MkuMul( _MkuValue *psV1, _MkuValue *psV2, _MkuValue *psR )
{
  char *pcV1, *pcV2, *pcI, *pcI_;
  int iL1, iL2, iS, iC;
  _MkuValue sI, sJ;

  pcV1 = psV1->pcAbs; iL1 = psV1->iLen;
  pcV2 = psV2->pcAbs; iL2 = psV2->iLen;

  _MkuParse( "0", &sJ );

  if( *pcV1 != '0' && *pcV2 != '0' )
  {
    pcI_ = sI.pcAbs + _BUFLEN - 1;
    *pcI_ = 0;

    while( iL2-- )
    {
      iL1 = psV1->iLen;
      pcI = pcI_;
      iC = 0;

      while( iL1-- )
      {
        iS = ( ( pcV1[iL1] - '0' ) * ( pcV2[iL2] - '0' ) + iC );
        iC = iS / 10;
        *--pcI = ( iS % 10 ) + '0';
      }

      if( iC )
        *--pcI = iC + '0';

      strcpy( sI.pcAbs, pcI );
      sI.iLen = strlen( sI.pcAbs );
      _MkuAdd( &sJ, &sI, &sJ );

      *--pcI_ = '0';
    }
  }

  *psR = sJ;
}

/* _MkuDiv
   division routine. a simple implementation of the "long integer" division
   algorithm that everyone had in school. not fast, but works.
   the function does not look at the signs of the two values, nor does
   it set the sign of the result. must be done by the caller.
*/
void _MkuDiv( _MkuValue *psV1, _MkuValue *psV2, _MkuValue *psR )
{
  char *pcV1, *pcV2, *pcR, pcBuf[_BUFLEN];
  int   i, iL1, iL2, iD;
  _MkuValue sV, sD;

  pcV1 = psV1->pcAbs; iL1 = psV1->iLen;
  pcV2 = psV2->pcAbs; iL2 = psV2->iLen;

  pcR = pcBuf;
  iD = 0;
  sV.iLen = 0;

  while( iD < iL1 )
  {
    for( ; iD < iL1; iD++, *pcR++ = '0' )
    {
      if( ! sV.iLen && pcV1[iD] == '0' )
        continue;

      sV.pcAbs[sV.iLen] = pcV1[iD];
      sV.pcAbs[++sV.iLen] = 0;

      if( _MkuCmpAbs( &sV, psV2 ) != -1 )
        break;
    }

    if( ! sV.iLen )
      break;

    iD++;

    _MkuParse( "9", &sD );
    _MkuMul( &sD, psV2, &sD );

    for( i = 9; _MkuCmpAbs( &sD, &sV ) == 1; i-- )
      _MkuSub( &sD, psV2, &sD );

    if( ! i )
      break;

    *pcR++ = i + '0';
    _MkuSub( &sV, &sD, &sV );

    if( sV.iLen == 1 && sV.pcAbs[0] == '0' )
      sV.iLen = 0;
  }

  *pcR = 0;
  for( pcR = pcBuf; *pcR == '0' && *(pcR+1); pcR++ );

  strcpy( psR->pcAbs, pcR );
  psR->iLen = strlen( psR->pcAbs );
}

/* _MkuExpr
   a simple expression parser. resolves expr's like "5 * ( 4 + 2 * 3 ) / 2".
   the expr must be a proper list, "5*(4+2*3)/3" is not allowed. searches from
   left to right for the first ")", then from there for the corresponding "("
   to find any "inner" expr, here "( 4 + 2 * 3 )". inside this inner expr, it
   searches for * and /, then + and -, then the comparison operators.
   for each found operator, the neighboring values are taken and this "atom"
   expr is calculated. then the atom expr is replaced by its result and
   the search for a ")" starts again. if a single value within "( )" is found,
   the "( )" is removed, and the inner expr is resolved.
   the example reduces to:
   ( 5 * ( 4 + 2 * 3 ) / 2 )   finds ( 4 + 2 * 3 )
   ( 5 * ( 4 + 6 ) / 2 )       finds the *, calc's 2 * 3, replaces it by 6
   ( 5 * ( 10 ) / 2 )          finds ( 10 )
   ( 5 * 10 / 2 )              removes the ( )
   ( 50 / 2 )                  finds the *, calc's 5 * 10, replaces it by 50
   ( 25 )                      finds the /, calc's 50 / 2, replaces it by 25
   25                          removes the ( ), done!

*/
int _MkuExpr( Tcl_Interp *pI, Tcl_Obj *poExpr )
{
  int     iL, iR, iO, iObjc;
  Tcl_Obj **ppoObjv, *poE, *poR;

  poE = _NSO( "(" );                  /* first, create an outermost () pair */
  try( _LOAL( poE, poExpr ), eError );              /* with the expr inside */
  try( _LOAE( poE, _NSO( ")" ) ), eError );    /* simplifies the algorithm! */

  while( 1 )
  {
    try( _LOGE( poE, &iObjc, &ppoObjv ), eError );

    if( iObjc == 1 )                           /* only one left, we're done */
      break;

                        /* now search for first ), then for corresponding ( */
    for( iR = 0 ; iR < iObjc && strcmp( _GSO( ppoObjv[iR] ), ")" ); iR++ );
    for( iL = iR; iL > 0     && strcmp( _GSO( ppoObjv[iL] ), "(" ); iL-- );

    if( iR - iL == 3 || iR == iObjc || iL == -1 )
      throw eBadSyntax;                       /* error, if ( or ) not found */

    if( iR - iL == 2 )                     /* just one value inside the ( ) */
    {
      try( _LODE( poE, iR ), eError );                    /* remove the ( ) */
      try( _LODE( poE, iL ), eError );
      continue;              /* inner expr is resolved, search for next ( ) */
    }

    for( iO = iL+1; iO < iR; iO++ )                    /* check for * and / */
      if( ! strcmp( _GSO( ppoObjv[iO] ), "*" ) || ! strcmp( _GSO( ppoObjv[iO] ), "/" ) )
        break;

    if( iO == iR )
      for( iO = iL+1; iO < iR; iO++ )                  /* check for + and - */
        if( ! strcmp( _GSO( ppoObjv[iO] ), "+" ) || ! strcmp( _GSO( ppoObjv[iO] ), "-" ) )
          break;

    if( iO == iR )
      for( iO = iL+1; iO < iR; iO++ )                /* check for comp. ops */
        if( ! strcmp( _GSO( ppoObjv[iO] ), "<"  ) || ! strcmp( _GSO( ppoObjv[iO] ), ">"  ) ||
            ! strcmp( _GSO( ppoObjv[iO] ), "<=" ) || ! strcmp( _GSO( ppoObjv[iO] ), ">=" ) ||
            ! strcmp( _GSO( ppoObjv[iO] ), "==" ) || ! strcmp( _GSO( ppoObjv[iO] ), "!=" ) )
          break;

    if( iO == iR )                       /* no valid operator found? error! */
      throw eBadSyntax;

    _ROR;                                    /* now calculate the atom expr */
    try( Mku_TulipCmd( NULL, pI, 4, &ppoObjv[iO-2] ), eBadSyntax );

    poR = _GOR;                           /* and replace it with its result */
    try( _LORR( poE, iO-1, iO+1, 1, &poR ), eError );
  }

  _SOR( poE );             /* the expr is now reduced to its overall result */

  return TCL_OK;

  catch eError:
    return TCL_ERROR;
  catch eBadSyntax:
    return _MkuError( pI, "Syntax error in expression '%O'", poExpr );
}

/* Mku_TulipCmd
   implements the tulip command. checks if either 1 or 3 arguments are given.
   for 1 argument it calls the parser function _MkuExpr to evaluate. for 3
   arguments, the middle one must be an allowed operator and the others must
   convert into _MkuValue structs. if the operator is a comparison operator,
   the _MkuCmp function provides the results. if its an arithmetic operator,
   the respective subfunctions are called, after the signs (+/-) of the two
   numbers are analyzed. an addition can actually be a substraction, if the
   signs are different. a substraction is actually an addition if the signs
   are not different (+x - -y = x + y, -x - +y = - (x + y) ).
   finally, the sign of the result is determined.
*/
int Mku_TulipCmd( ClientData pC, Tcl_Interp *pI, int objc, Tcl_Obj *CONST objv[] )
{
  int        iOp, bR, iCmp;
  _MkuValue  sV1, sV2, sR;

  static char *ppcOps[] = { "+" , "-" , "*" , "/" , "<", ">", "<=", ">=", "==", "!=", NULL };
  static enum  ops        { _ADD, _SUB, _MUL, _DIV, _LT, _GT, _LE , _GE , _EQ , _NE };

  if( objc != 2 && objc != 4 )
    return _WNA( 1, "expr OR tulip value1 operand value2" );

  if( objc == 2 )
  {
    try( _MkuExpr( pI, objv[1] ), eError );       /* analyze the expression */
  }
  else
  {
    try( _GIFO( objv[2], ppcOps, "operand", &iOp ), eError );

    if( _MkuParse( _GSO( objv[1] ), &sV1 ) )
      throw eBadValue1;
    else if( _MkuParse( _GSO( objv[3] ), &sV2 ) )
      throw eBadValue2;

    if( iOp <= _DIV )                               /* arithmetic operators */
    {
      switch( iOp )
      {
        case _ADD:
        case _SUB:
          if( ( iOp == _ADD ) ^ ( sV1.bNeg != sV2.bNeg ) )       /* add, if */
          {              /* ADD and signs are same, or SUB and signs differ */
            _MkuAdd( &sV1, &sV2, &sR );
            sR.bNeg = sV1.bNeg;
          }
          else                              /* it's actually a substraction */
          {
            iCmp = _MkuCmpAbs( &sV1, &sV2 );
            sR.bNeg = ( sV1.bNeg ^ ( iCmp == -1 ) );

            if( iCmp == -1 )       /* make sure lesser value is substracted */
              _MkuSub( &sV2, &sV1, &sR );               /* from the greater */
            else
              _MkuSub( &sV1, &sV2, &sR );
          }

          break;

        case _MUL:
          if( *sV1.pcAbs == '0' || *sV2.pcAbs == '0' )   /* result is 0, if */
            _MkuParse( "0", &sR );              /* one of the operands is 0 */
          else
          {
            _MkuMul( &sV1, &sV2, &sR );
            sR.bNeg = ( sV1.bNeg != sV2.bNeg );
          }

          break;

        case _DIV:
          if( *sV2.pcAbs == '0' )                 /* catch division by zero */
            throw eDivByZero;

          iCmp = _MkuCmpAbs( &sV1, &sV2 );
          if( iCmp == 0 )
            _MkuParse( "1", &sR );     /* result is 1 if operands are equal */
          else if( iCmp == -1 )
            _MkuParse( "0", &sR );          /* result is 0 if |op1| < |op2| */
          else
          {
            _MkuDiv( &sV1, &sV2, &sR );
            sR.bNeg = ( sV1.bNeg != sV2.bNeg );
          }

          break;
      }

      if( sR.bNeg && *sR.pcAbs != '0' )
        _ASO( _GOR, "-", sR.pcAbs, NULL );
      else
        _ASO( _GOR, sR.pcAbs, NULL);
    }
    else                                            /* comparison operators */
    {
      switch( iOp )
      {
        case _LT: bR = ( _MkuCmp( &sV1, &sV2 ) == -1 ); break;
        case _GT: bR = ( _MkuCmp( &sV1, &sV2 ) == 1 ) ; break;
        case _LE: bR = ( _MkuCmp( &sV1, &sV2 ) <= 0 ) ; break;
        case _GE: bR = ( _MkuCmp( &sV1, &sV2 ) >= 0 ) ; break;
        case _EQ: bR = ( _MkuCmp( &sV1, &sV2 ) == 0 ) ; break;
        case _NE: bR = ( _MkuCmp( &sV1, &sV2 ) != 0 ) ; break;
      }

      _SOR( _NBO( bR ) );
    }
  }

  return TCL_OK;

  catch eError:
    return TCL_ERROR;
  catch eBadValue1:
    return _MkuError( pI, "Format error in value '%O'", objv[1] );
  catch eBadValue2:
    return _MkuError( pI, "Format error in value '%O'", objv[3] );
  catch eDivByZero:
    return _MkuError( pI, "Division by zero" );
}

/* Mktulip_Init, Mktulip_SafeInit
   package initialization. creates the tulip command and registers the
   package.
*/
int Mktulip_Init( Tcl_Interp *pI )
{
#ifdef USE_TCL_STUBS
  if( Tcl_InitStubs( pI, "8.2", 0) == NULL )
    throw eError;
#endif

  if( TCL_MAJOR_VERSION < 8 )
    throw eWrongVersion;

  Tcl_CreateObjCommand( pI, "tulip" , Mku_TulipCmd , NULL, NULL );

  try( Tcl_PkgProvide( pI, "mkTulip", _VERSION ), eError );

  return TCL_OK;

  catch eError:
    return TCL_ERROR;

  catch eWrongVersion:
    _SSO( _GOR, "Package mkTulip requires Tcl Version 8" );
    return TCL_ERROR;
}

int Mktulip_SafeInit( Tcl_Interp *pI )
{
  return Mktulip_Init( pI );
}

/* static linking. uncomment the following two functions if you want
   to create a stand-alone shell instead of a dynamic library. */

#ifndef USE_TCL_STUBS
int main( int argc, char *argv[] )
{
  Tcl_Main( argc, argv, Tcl_AppInit );
  return 0;
}

int Tcl_AppInit( Tcl_Interp *pI )
{
  try( Tcl_Init( pI ), eError );
  try( Mktulip_Init( pI ), eError );

  return TCL_OK;

  catch eError:
    return TCL_ERROR;
}
#endif

/*
 * mkTulip 1.0
 * -----------
 */

